from asyncio import Event, sleep
from datetime import datetime
from typing import Optional

import pytz
from environs import Env
from pytdx.exhq import TdxExHq_API, TDXParams
from pytdx.parser.ex_get_instrument_bars import GetInstrumentBars
from pytz import timezone
from livetrader.exceptions import RemoteError
from livetrader.market import MarketBase


class TdxMarket(MarketBase):
    __market_name__ = 'TDX'
    __timeframe__ = '1MIN'

    def __init__(self, host: Optional[str] = None):
        self._env = env = Env()
        env.read_env()
        tdx_host = host if host else env.str('TDX_HOST')
        self._api = TdxExHq_API(heartbeat=True, multithread=True)
        self._ip, self._port = tdx_host.split(':')
        self._server_tz = timezone('Asia/Shanghai')
        self._pill = Event()
        self._market_mapping = dict(
            zip(["SHFE", "CZCE", "DCE", "CFFEX", "US"], [30, 28, 29, 47, 74]))

    def connect(self):
        if not self._api.connect(self._ip, int(self._port)):
            raise RemoteError('%s:%s connect error.' %
                              (self.ip, int(self.port)))
        self._market_list = list(
            self._api.to_df(
                self._api.get_markets()).market)

    async def watch_klines(self, symbol: str):
        market, code = symbol.split('.')
        last_kline = None
        while not self._pill.is_set():
            tdx_klines = self._get_instrument_bars(
                TDXParams.KLINE_TYPE_1MIN, self._market_mapping[market], code, 0, 10)
            # 这个地方处理的原因在于由于是轮询的，就有可能出现在轮询间隔的时间(这里是1秒)内，有一半时间属于上一根K线，有一半时间属于下一根K线，所以如果我们只看最新的K线的话，前一根K线的一部分数据可能就会丢失，因此这里判断一下前一根K线如果有更新，就先把前一根K线推送了，再推送下一根K线
            prev_kline = self._parse_kline(tdx_klines[-2])
            latest_kline = self._parse_kline(tdx_klines[-1])
            if last_kline and prev_kline['datetime'] == last_kline['datetime']:
                self.logger.debug('yield prev kline: %s' % prev_kline)
                yield prev_kline
            if last_kline != latest_kline:
                yield latest_kline
                self.logger.debug('yield latest kline: %s' % latest_kline)
                last_kline = latest_kline
            await sleep(1)

    async def get_kline_histories(self, symbol: str, from_ts: Optional[int] = None, limit: Optional[int] = None):
        market, code = symbol.split('.')
        if self._market_mapping[market] in self._market_list:
            bars = []
            if from_ts is not None:
                idx = 0
                while len(bars) == 0 or bars[-1]['datetime'] >= from_ts:
                    bars += [
                        self._parse_kline(bar) for bar in reversed(
                            self._get_instrument_bars(
                                TDXParams.KLINE_TYPE_1MIN,
                                self._market_mapping[market],
                                code, idx * 700, 700))]
                    idx += 1
                # 前面都是整700地取，所以有可能取到的数据会超过 from_ts,因此这里再做一次过滤
                bars = list(filter(lambda x: x['datetime'] >= from_ts, bars))
            elif limit is not None:
                for i in range(limit // 700):
                    bars += [
                        self._parse_kline(bar) for bar in reversed(
                            self._get_instrument_bars(
                                TDXParams.KLINE_TYPE_1MIN,
                                self._market_mapping[market],
                                code, i * 700, 700))]
                bars += [
                    self._parse_kline(bar) for bar in reversed(
                        self._get_instrument_bars(
                            TDXParams.KLINE_TYPE_1MIN,
                            self._market_mapping[market],
                            code, limit // 700 * 700,
                            limit % 700))]
            return reversed(bars)

    def disconnect(self):
        self._pill.set()
        self._api.disconnect()

    def _get_instrument_bars(self, category, market, code, start=0, count=700):
        while True:
            try:
                cmd = GetInstrumentBars(self._api.client, lock=self._api.lock)
                cmd.setParams(category, market, code, start=start, count=count)
                return cmd.call_api()
            except Exception as e:
                pass

    def _parse_kline(self, kline_tdx: dict) -> dict:
        dt = self._server_tz.localize(
            datetime.strptime(
                kline_tdx['datetime'],
                '%Y-%m-%d %H:%M')).astimezone(pytz.utc)
        return {
            'datetime': int(dt.timestamp() * 1000),
            'open': str(round(kline_tdx['open'], 2)),
            'high': str(round(kline_tdx['high'], 2)),
            'low': str(round(kline_tdx['low'], 2)),
            'close': str(round(kline_tdx['close'], 2)),
            'volume': kline_tdx['trade'],
        }




from typing import List, Optional

from livetrader.rpc import Method, Publisher, Subscriber


# 仅用来标识一些常量，具体order对象依旧是 dict 对象
class OrderBase():
    ExecTypes = ['Market', 'Close', 'Limit', 'Stop', 'StopLimit', 'StopTrail',
                 'StopTrailLimit', 'Historical']

    OrdTypes = ['Buy', 'Sell']
    Buy, Sell = range(2)

    Created, Submitted, Accepted, Partial, Completed, \
        Canceled, Expired, Rejected = range(9)

    Cancelled = Canceled  # alias

    Status = [
        'Created', 'Submitted', 'Accepted', 'Partial', 'Completed',
        'Canceled', 'Expired', 'Rejected',
    ]


# 为了适配不同的交易接口，有些是推送数据（比方说期货）有些是拉数据（比方说数字货币），我们都用TradeBase 来代表，然后用 TradeService 封装出统一接口
# 在 TradeBase 里如果是推送的接口，就直接 yield 出来，如果是拉数据的，就定时 polling，然后缓存在
# TradeService 里（数据库），之后都统一出推送接口和polling 接口，这样TradeBase可以自己控制调用接口的频率，然后
# TradeService 对外就没有调用限制了


class TradeBase():
    def connect(self):
        raise NotImplementedError()

    def disconnect(self):
        raise NotImplementedError()

    async def fetch_balance(self):
        """
        balance 里包含字段：
            cash：现金
            value: 总资产
        """
        raise NotImplementedError()

    async def fetch_markets(self):
        """
        获取可交易品种信息
            symbol 品种
            leverage 杠杆大小
            mult 乘数
            commission 费率
            margin 保证金
            shortable 可做空
        """
        raise NotImplementedError()

    async def create_order(self, symbol: str, ordtype: int, exectype: int, price: float, size: int, client: Optional[str] = None):
        """
        symbol: 标的，SHFE.rb2101
        ordtype: OrderBase.OrdTypes
        exectype: OrderBase.ExecTypes
        price: 价格
        size: 正数为 Long 多单, 负数为 Short 空单
        client: 客户端的id，调用用来区分是否为自己发起的订单（与成交）

        返回：多一个 status 属性， OrderBase.Status
        因为不是所有的市场关于定单都是阻塞的，就哪怕是市价单也会通过异步的方式通知订单状态，所以这里定单会有很多不同的状态，参考 OrderBase.Status
        """
        raise NotImplementedError()

    async def cancel_order(self, order_id: str):
        """
        这里和 create_order 一样，也是异步调用，返回只能代表交易所收到这个请求，不一定执行成功，有可能会被拒绝，需要 watch_orders 看结果
        """
        raise NotImplementedError()

    async def watch_orders(self, symbol: str):
        raise NotImplementedError()

    async def fetch_orders(self, symbol: Optional[str] = None):
        raise NotImplementedError()

    async def fetch_open_orders(self, symbol: Optional[str] = None):
        raise NotImplementedError()

    async def fetch_trades(self, symbol: Optional[str] = None):
        raise NotImplementedError()

    async def watch_trades(self, symbol: str):
        raise NotImplementedError()

    async def fetch_positions(self, symbol: Optional[str] = None):
        raise NotImplementedError()


class TradeSubscriber(Subscriber):

    def on_order(self, order: dict):
        raise NotImplementedError()

    def on_trade(self, trade: dict):
        raise NotImplementedError()


class TradeService():

    def __init__(self):
        self._publishers = []
        self._tasks = []

    async def _publish(self):
        raise NotImplementedError()

    def start(self):
        # start publish service(polling and push)
        raise NotImplementedError()

    @Method
    async def get_balance(self):
        raise NotImplementedError()

    @Method
    async def create_order(self, symbol: str, exectype: str, price: float, size: int):
        raise NotImplementedError()

    @Method
    async def cancel_order(self, order_id: str):
        raise NotImplementedError()

    @Method
    async def open_orders(self, symbol: Optional[str] = None):
        raise NotImplementedError()

    @Method
    async def today_trades(self, symbol: Optional[str] = None):
        raise NotImplementedError()

    @Method
    async def fetch_position(self, symbol: Optional[str] = None):
        raise NotImplementedError()

    def stop(self):
        for publisher in self._publishers:
            publisher.close()
        for task in self._tasks:
            task.cancel()



import asyncio
import functools
import logging
import signal

import gevent
import zerorpc
from zerorpc import Client as _Client
from zerorpc import Publisher as _Publisher
from zerorpc import Subscriber as _Subscriber

from livetrader.utils import FifoQueue


class Publisher(_Publisher):
    pass


class Subscriber(_Subscriber):
    pass


class Client(_Client):
    pass


class MarketSubscriber(Subscriber):

    def __init__(self, symbol: str):
        super().__init__()
        self._symbol = symbol
        self._pill2kill = asyncio.Event()

    def connect(self, endpoint, resolve=True):
        _endpoint = "%s_%s" % (endpoint, self._symbol)
        return super().connect(endpoint=_endpoint, resolve=resolve)

    def on_kline(self, kline: dict):
        raise NotImplementedError()

    def run(self):
        self._gevent_task = gevent.spawn(super().run)

        async def the_loop():
            while not self._pill2kill.is_set():
                await asyncio.sleep(0)
                gevent.sleep(0)
        self._asyncio_task = asyncio.get_event_loop().create_task(the_loop())

    def close(self):
        self._pill2kill.set()
        if self._gevent_task is not None:
            self._gevent_task.kill()
        if self._asyncio_task is not None:
            self._asyncio_task.cancel()


class Method(object):
    def __init__(self, functor):
        self._functor = functor
        self.__doc__ = functor.__doc__
        self.__name__ = getattr(functor, "__name__", str(functor))
        functools.update_wrapper(self, functor)

    @property
    def coroutine(self):
        return asyncio.iscoroutinefunction(self._functor)

    def __get__(self, instance, type_instance=None):
        if instance is None:
            return self
        return self.__class__(self._functor.__get__(instance, type_instance))

    def __call__(self, *args, **kwargs):
        return self._functor(*args, **kwargs)


class Server(object):

    def __init__(self, service):
        self._service = service
        self._publishers = {}
        self._logger = logging.getLogger('Server')
        methods = dict((k, self._decorate_coroutine_method(getattr(service, k))) for k in dir(service)
                       if isinstance(getattr(service, k), Method))
        self.s = zerorpc.Server(methods=methods)
        self._pill2kill = asyncio.Event()

    def bind(self, endpoint: str):
        self._endpoint = endpoint
        self.s.bind(endpoint)

    def _decorate_coroutine_method(self, method):
        if method.coroutine:
            def __deco__(*args, **kwargs):
                return asyncio.get_event_loop().run_until_complete(method(*args, **kwargs))
            return __deco__
        else:
            return method

    def _publish(self, queue: FifoQueue):
        while not self._pill2kill.is_set():
            gevent.sleep(0)
            if not queue.empty():
                symbol, kline = queue.get_nowait()
                if symbol in self._publishers:
                    publisher = self._publishers[symbol]
                else:
                    publisher = self._publishers[symbol] = Publisher()
                    publisher.bind('%s_%s' % (self._endpoint, symbol))
                publisher.on_kline(kline)

    def run(self):
        # register shutdown handler
        gevent.signal_handler(signal.SIGINT, self.close)
        gevent.signal_handler(signal.SIGTERM, self.close)

        queue = self._service.start()
        self._publish_task = gevent.spawn(self._publish, (queue))
        self._server_task = gevent.spawn(self.s.run)

        self._logger.info('Server started')
        while not self._pill2kill.is_set():
            asyncio.get_event_loop().run_until_complete(asyncio.sleep(0))
            gevent.sleep(0)

    def close(self):
        self._logger.info('Server stopping...')
        if self._publish_task is not None:
            self._publish_task.kill()
        self.s.close()
        if self._server_task is not None:
            self._server_task.kill()
        for publisher in self._publishers.values():
            publisher.close()
        self._service.stop()
        self._logger.info('Server stopped')
        self._pill2kill.set()




import asyncio


class FifoQueue(asyncio.Queue):

    async def put(self, item):
        if self.maxsize > 0 and self.qsize() >= self.maxsize:
            await super().get()
        await super().put(item)

    def put_nowait(self, item):
        if self.maxsize > 0 and self.qsize() >= self.maxsize:
            super().get_nowait()
        return super().put_nowait(item)



import asyncio
from asyncio import Queue, sleep
from datetime import datetime
from typing import Optional

import pytz
import tzlocal
from environs import Env
from pandas import Timedelta, Timestamp
from pytz import timezone
from livetrader.lib.dwx_zeromq_connector import DWX_ZeroMQ_Connector
from livetrader.market import MarketBase


class DwxMarket(MarketBase):
    __market_name__ = 'DWX'
    __timeframe__ = '1MIN'

    def __init__(self, host: Optional[str] = None,
                 push_port: Optional[int] = None,
                 pull_port: Optional[int] = None,
                 sub_port: Optional[int] = None,
                 time_zone: Optional[str] = None):
        self._env = env = Env()
        env.read_env()
        self._host = host if host else env.str('DWX_HOST')
        self._push_port = push_port if push_port else env.int(
            'DWX_PUSH_PORT', 32768)
        self._pulling_port = pull_port if pull_port else env.int(
            'DWX_PULL_PORT', 32769)
        self._subscribe_port = sub_port if sub_port else env.int(
            'DWX_SUB_PORT', 32770)
        self._server_tz = timezone(
            time_zone if time_zone else env.str(
                'DWX_TIMEZONE', 'EET'))

    def connect(self):
        self._kline_resp = {}
        self._kline_sub = {}
        self._connector = DWX_ZeroMQ_Connector(
            _host=self._host,
            _PUSH_PORT=self._push_port,
            _PULL_PORT=self._pulling_port,
            _SUB_PORT=self._subscribe_port,
            _subdata_handlers=[self],
            _pulldata_handlers=[self])

    def disconnect(self):
        self._connector._DWX_ZMQ_SHUTDOWN_()

    async def onSubData(self, msg):
        code, rates = msg.split(' ')
        code = code.split('_')[0]
        time, open,  high,  low,  close,  tick_volume, spread, real_volume = rates.split(
            ';')
        dt = self._server_tz.localize(
            datetime.utcfromtimestamp(int(time))).astimezone(pytz.utc)
        kline = {
            'datetime': int(dt.timestamp()) * 1000,
            'open': open,
            'high': high,
            'low': low,
            'close': close,
            'spread': spread,
            'volume': int(tick_volume)
        }
        await self._kline_sub[code].put(kline)

    async def onPullData(self, msg):
        if msg['_action'] and msg['_action'] == 'HIST':
            queue = self._kline_resp[msg['_symbol'].split('_')[0]]
            await queue.put(msg['_data'])

    async def watch_klines(self, symbol: str):
        market, code = symbol.split('.')
        if market == 'FOREX':
            self._kline_sub[code] = Queue()
            await self._connector._DWX_MTX_SUBSCRIBE_MARKETDATA_(code)
            await self._connector._DWX_MTX_SEND_TRACKRATES_REQUEST_(
                [('%s_M1' % code, code, 1)])
            while True:
                kline = await self._kline_sub[code].get()
                yield kline

    async def get_kline_histories(self, symbol: str, from_ts: Optional[int] = None, limit: Optional[int] = None):
        market, code = symbol.split('.')
        if market == 'FOREX':
            local_tz = tzlocal.get_localzone()
            _end = local_tz.localize(Timestamp.now())
            if from_ts:
                _start = local_tz.localize(Timestamp.fromtimestamp(
                    from_ts / 1000))
            else:
                _start = _end - Timedelta(minutes=limit - 1)
            if _start >= _end:
                return []
            _start = _start.astimezone(self._server_tz)
            _end = _end.astimezone(self._server_tz)
            self._kline_resp[code] = Queue()
            await self._connector._DWX_MTX_SEND_HIST_REQUEST_(_symbol=code,
                                                              _timeframe=1,
                                                              _start=_start.strftime(
                                                                  '%Y.%m.%d %H:%M:00'),
                                                              _end=_end.strftime('%Y.%m.%d %H:%M:00'))
            klines = [self._parse_kline(kline_raw) for kline_raw in await self._kline_resp[code].get()]
            del self._kline_resp[code]
            return klines
        else:
            return []

    def _parse_kline(self, kline_raw: dict) -> dict:
        dt = self._server_tz.localize(
            datetime.strptime(
                kline_raw['time'],
                '%Y.%m.%d %H:%M')).astimezone(
            pytz.utc)
        return {
            'datetime': int(dt.timestamp() * 1000),
            'open': kline_raw['open'],
            'high': kline_raw['high'],
            'low': kline_raw['low'],
            'close': kline_raw['close'],
            'spread': kline_raw['spread'],
            'volume': kline_raw['tick_volume']
        }



import asyncio
import logging
from datetime import datetime
from typing import List, Optional

from livetrader.rpc import Method, Publisher
from livetrader.utils import FifoQueue


class MarketBase(object):

    __market_name__ = None
    __timeframe__ = None

    def connect(self):
        raise NotImplementedError()

    def disconnect(self):
        raise NotImplementedError()

    @property
    def logger(self):
        return logging.getLogger(self.__class__.__name__)

    async def watch_klines(self, symbol: str):
        raise NotImplementedError()

    async def get_kline_histories(self, symbol: str, from_ts: Optional[int] = None, limit: Optional[int] = None):
        raise NotImplementedError()


class MarketService(object):

    def __init__(self, market: MarketBase, symbols: List[str]):
        self._market = market
        self._symbols = symbols
        self._tasks = []
        self.__logger__ = logging.getLogger('MarketService')

    async def _publish(self, symbol: str, queue: FifoQueue):
        async for kline in self._market.watch_klines(symbol):
            await queue.put((symbol, kline))

    def start(self):
        queue = FifoQueue(maxsize=100)
        self._market.connect()
        for symbol in self._symbols:
            self._tasks.append(asyncio.get_event_loop().create_task(
                self._publish(symbol, queue)))
        return queue

    @Method
    async def get_kline_histories(
            self, symbol: str, from_ts: Optional[datetime] = None, limit: Optional[int] = None):
        return list(await self._market.get_kline_histories(symbol, from_ts, limit))

    def stop(self):
        for task in self._tasks:
            task.cancel()
        self._market.disconnect()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="livetrader",  # Replace with your own username
    version="0.1",
    author='lookis<Jingsi Liu>',
    author_email="lookisliu@gmail.com",
    description="A small and extendable framework for trading",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lookis/livetrader",
    packages=setuptools.find_packages(),
    install_requires=[
        'click',
        'tzlocal',
        'environs',
        'zerorpc',
        'pymongo',
        'motor',
    ],
    python_requires='>=3.6',
)



from typing import List, Optional

from environs import Env
from future.utils import iteritems
from motor.motor_asyncio import AsyncIOMotorClient
from pymongo import ReplaceOne
from livetrader.market import MarketBase


class CachedMarket(MarketBase):

    def __init__(self, market: MarketBase, mongodb_uri: Optional[str] = None):
        self._env = env = Env()
        env.read_env()
        self._mongodb_uri = mongodb_uri if mongodb_uri else env.str(
            'MONGODB_URI')
        self._market = market
        self._mongo_client = AsyncIOMotorClient(self._mongodb_uri)
        self._database = self._mongo_client.get_default_database('markets')
        self._hist_mongo_client = None
        self._initied = False

    def connect(self):
        self._market.connect()

    def disconnect(self):
        self._mongo_client.close()
        self._market.disconnect()

    def _collection(self, symbol: str):
        db = self._database
        if self._market.__class__.__market_name__ is None or self._market.__class__.__timeframe__ is None:
            raise Exception(
                'Initialize market cache error (without market_name or timeframe)')
        collection_name = '%s_%s_%s' % (
            self._market.__class__.__market_name__, symbol, self._market.__class__.__timeframe__)
        collection = db[collection_name]
        return collection

    async def _insert_db(self, symbol: str, klines: List[dict]):
        collection = self._collection(symbol)
        await collection.bulk_write([
            ReplaceOne({'datetime': kline['datetime']}, kline, upsert=True)
            for kline in klines
        ])

    async def _initial_cache(self, symbol: str):
        collection = self._collection(symbol)
        if await collection.count_documents({}) > 0:
            rs = await (collection.find().sort([('datetime', -1)]).limit(1).to_list(1))
            last_kline = rs[0]
            history_klines = await self._market.get_kline_histories(
                symbol, from_ts=last_kline['datetime'])
        else:
            await collection.create_index('datetime', unique=True)
            history_klines = await self._market.get_kline_histories(
                symbol, limit=5000)
        await self._insert_db(symbol, history_klines)

    async def watch_klines(self, symbol: str):
        if not self._initied:
            self.logger.debug('init cache from watch_klines')
            await self._initial_cache(symbol)
            self._initied = True
        self.logger.debug('do watch klines from delegate')
        async for kline in self._market.watch_klines(symbol):
            await self._insert_db(symbol, [kline])
            yield kline

    async def get_kline_histories(self, symbol: str, from_ts: Optional[int] = None, limit: Optional[int] = None):
        if not self._initied:
            await self._initial_cache(symbol)
            self._initied = True
        collection = self._collection(symbol)
        criteria = {}
        if from_ts:
            criteria = {'datetime': {'$gte': from_ts}}
        cursor = collection.find(criteria).sort([('datetime', -1)])
        if limit:
            cursor = cursor.limit(limit)
            count = limit
        else:
            count = await collection.count_documents(criteria)
        return reversed(list(map(lambda x: {k: v for k, v in iteritems(x) if not k.startswith('_')}, await cursor.to_list(count))))
